/*-
 * ~~~~~~licensing~~~~~~
 * application-shell
 * ==========
 * Copyright (C) 2020 - 2025 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.integration.applications.shell.application;

import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

import org.apache.karaf.features.FeaturesService;
import org.apache.karaf.shell.api.action.Command;
import org.apache.karaf.shell.api.action.lifecycle.Reference;
import org.apache.karaf.shell.api.action.lifecycle.Service;

import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.support.karaf.shell.ShellTableFixed;
import ru.entaxy.platform.base.support.osgi.bundle.CapabilityDescriptor;
import ru.entaxy.platform.integration.applications.EntaxyApplication;
import ru.entaxy.platform.integration.applications.EntaxyApplicationVersion;
import ru.entaxy.platform.integration.applications.shell.ApplicationManagerCommandSupport;
import ru.entaxy.platform.integration.applications.shell.ApplicationShell;
import ru.entaxy.platform.integration.applications.shell.FeatureCapabilityHelper;

@Service
@Command(name = "app-list", scope = ApplicationShell.COMMAND_SCOPE)
public class ApplicationListCommand extends ApplicationManagerCommandSupport {

    protected final static String APP_NAMESPACE = "entaxy.application";

    @Reference
    FeaturesService featuresService;

    @Override
    public Object execute() throws Exception {
        ShellTableFixed table = new ShellTableFixed();
        table.column("Name");
        table.column("Storage");
        table.column("Versions");
        table.column("Last revision").alignCenter();
        table.column("Installed revision").alignCenter();

        List<EntaxyApplication> apps = applicationManager.getApplications();
        Collections.sort(apps, new Comparator<EntaxyApplication>() {

            @Override
            public int compare(EntaxyApplication o1, EntaxyApplication o2) {
                String storage1 = o1.getStorage().getName();
                String storage2 = o2.getStorage().getName();
                if (!storage1.equals(storage2))
                    return storage1.compareTo(storage2);

                return o1.getName().compareTo(o2.getName());
            }

        });

        for (EntaxyApplication app : apps) {
            List<EntaxyApplicationVersion> versions = app.getVersions();
            Collections.sort(versions, new Comparator<EntaxyApplicationVersion>() {

                @Override
                public int compare(EntaxyApplicationVersion o1, EntaxyApplicationVersion o2) {
                    return o1.compareTo(o2);
                }

            });
            String versionsValue = versions.stream().map(v -> v.getVersionNumber()).collect(Collectors.joining("\n"));
            String lastRevisionsValue = versions.stream().map(v -> v.getLastRevision().getRevisionNumber() + "")
                    .collect(Collectors.joining("\n"));
            String installedRevisionsValue = versions.stream().map(v -> getInstalledRevisionValue(v))
                    .collect(Collectors.joining("\n"));;

            table.addRow().addContent(app.getName(), app.getStorage() == null ? "" : app.getStorage().getName(),
                    versionsValue, lastRevisionsValue, installedRevisionsValue);
        }

        // print output
        table.print(System.out);

        return null;
    }

    protected String getInstalledRevisionValue(EntaxyApplicationVersion version) {
        try {
            List<FeatureCapabilityHelper> helpers = Arrays.asList(featuresService.listInstalledFeatures()).stream()
                    .map(f -> new FeatureCapabilityHelper(f))
                    .filter(h -> h.isCapabilityProvided(APP_NAMESPACE))
                    .collect(Collectors.toList());

            for (FeatureCapabilityHelper helper : helpers) {
                List<CapabilityDescriptor> descriptors = helper.getProvidedCapabilities(APP_NAMESPACE);
                for (CapabilityDescriptor descriptor : descriptors) {
                    String nameValue = descriptor.getAttributes().getOrDefault("name", "").toString();
                    String versionValue = descriptor.getAttributes().getOrDefault("version", "").toString();
                    if (version.getApplication().getName().equals(nameValue)
                            && version.getVersionNumber().equals(versionValue)) {
                        String revision = descriptor.getAttributes().getOrDefault("revision", "").toString();
                        if (CommonUtils.isValid(revision))
                            return revision;
                    }

                }

            }

        } catch (Exception ignore) {
            // NOOP
            ignore.printStackTrace();
        }
        return "";
    }

}
