/*-
 * ~~~~~~licensing~~~~~~
 * artifact-management-extensions
 * ==========
 * Copyright (C) 2020 - 2025 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.artifact.ext.features.impl;

import java.util.ArrayList;
import java.util.List;

import org.apache.karaf.cellar.features.management.CellarFeaturesMBean;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.support.osgi.OSGIUtils;
import ru.entaxy.platform.core.artifact.ext.features.impl.FeatureInstallerImpl.FeatureInstallDescriptorImpl;
import ru.entaxy.platform.core.artifact.ext.impl.ClusterTypedConfigurableHelper;
import ru.entaxy.platform.core.artifact.installer.builder.impl.InstallationResultImpl;
import ru.entaxy.platform.runtime.cellar.sequence.CellarSequenceManager;

public class FeatureInstallerClusterHelper extends ClusterTypedConfigurableHelper<FeatureInstallerImpl>
        implements FeatureInstallerConfigurableHelper {

    private static final Logger LOG = LoggerFactory.getLogger(FeatureInstallerClusterHelper.class);

    protected CellarFeaturesMBean getFeaturesService() throws Exception {
        if (CommonUtils.isValid(config.sequenceId)) {
            CellarSequenceManager seqManager =
                    OSGIUtils.services().ofClass(CellarSequenceManager.class).waitService(2000).get();

            return seqManager.getSequencedHelper(config.sequenceId, CellarFeaturesMBean.class);

        }
        return OSGIUtils.services().ofClass(CellarFeaturesMBean.class).waitService(2000).get();
    }

    @Override
    public InstallationResultImpl install(InstallationResultImpl installationResult) {
        CellarFeaturesMBean featuresService;
        try {
            featuresService = getFeaturesService();
        } catch (Exception e) {
            LOG.error("Error getting CellarFeaturesMBean service", e);
            return installationResult.failed("Error getting CellarFeaturesMBean service").error(e);
        }

        List<String> groupsToInstall = new ArrayList<>();
        for (String groupName : getGroups())
            if (checkGroup(groupName))
                if (!groupsToInstall.contains(groupName))
                    groupsToInstall.add(groupName);

        if (groupsToInstall.size() == 0) {
            return installationResult.failed("No groups to install to");
        }

        // TODO decide if we really need several groups
        // now we use only the first one
        String groupToInstall = groupsToInstall.get(0);

        boolean hasFailed = false;
        boolean hasSuccessfull = false;


        for (FeatureInstallDescriptorImpl descriptor : config.descriptors.values()) {
            InstallationResultImpl subResult = InstallationResultImpl.create();
            subResult.object(descriptor.featureName);
            try {

                // convert version to OSGI-style version
                // mj.mn.p-q -> mj.mn.p.q
                String finalFeatureVersion = descriptor.featureVersion;
                if (CommonUtils.isValid(finalFeatureVersion)) {
                    finalFeatureVersion = finalFeatureVersion.replace('-', '.');
                }

                featuresService.installFeature(groupToInstall, descriptor.featureName,
                        finalFeatureVersion,
                        descriptor.noRefresh,
                        descriptor.noStart,
                        descriptor.noManage,
                        descriptor.upgrade);
                subResult.installed();
                hasSuccessfull = true;
            } catch (Exception e) {
                LOG.error("Error installing repository [" + descriptor.featureName + "]", e);
                subResult.failed("Error installing feature [" + descriptor.featureName + "]").error(e);
                hasFailed = true;
            }
            installationResult.subResult(subResult);
        }

        if (!hasFailed)
            installationResult.installed();
        else {
            if (hasSuccessfull) {
                installationResult.message("HAS FAILED SUBRESULTS");
                installationResult.installed();
            } else {
                installationResult.failed();
            }
        }
        return installationResult;
    }

    @Override
    public InstallationResultImpl uninstall(InstallationResultImpl installationResult) {
        CellarFeaturesMBean featuresService;
        try {
            featuresService = getFeaturesService();
        } catch (Exception e) {
            LOG.error("Error getting CellarFeaturesMBean service", e);
            return installationResult.failed("Error getting CellarFeaturesMBean service").error(e);
        }

        List<String> groupsToUninstall = new ArrayList<>();
        for (String groupName : getGroups())
            if (checkGroup(groupName))
                if (!groupsToUninstall.contains(groupName))
                    groupsToUninstall.add(groupName);

        if (groupsToUninstall.size() == 0) {
            return installationResult.failed("No groups to uninstall from");
        }

        // TODO decide if we really need several groups
        // now we use only the first one
        String groupToUninstall = groupsToUninstall.get(0);

        boolean hasFailed = false;
        boolean hasSuccessfull = false;


        for (FeatureInstallDescriptorImpl descriptor : config.descriptors.values()) {
            InstallationResultImpl subResult = InstallationResultImpl.create();
            subResult.object(descriptor.featureName);
            try {

                // convert version to OSGI-style version
                // mj.mn.p-q -> mj.mn.p.q
                String finalFeatureVersion = descriptor.featureVersion;
                if (CommonUtils.isValid(finalFeatureVersion)) {
                    finalFeatureVersion = finalFeatureVersion.replace('-', '.');
                }

                featuresService.uninstallFeature(groupToUninstall, descriptor.featureName, finalFeatureVersion,
                        descriptor.noRefresh);
                subResult.uninstalled();
                hasSuccessfull = true;
            } catch (Exception e) {
                LOG.error("Error uninstalling repository [" + descriptor.featureName + "]", e);
                subResult.failed("Error uninstalling feature [" + descriptor.featureName + "]").error(e);
                hasFailed = true;
            }
            installationResult.subResult(subResult);
        }

        if (!hasFailed)
            installationResult.uninstalled();
        else {
            if (hasSuccessfull) {
                installationResult.message("HAS FAILED SUBRESULTS");
                installationResult.uninstalled();
            } else {
                installationResult.failed();
            }
        }
        return installationResult;
    }

}
