/*-
 * ~~~~~~licensing~~~~~~
 * object-management
 * ==========
 * Copyright (C) 2020 - 2025 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.management.object.impl;

import java.util.Dictionary;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import javax.management.DynamicMBean;
import javax.management.MBeanRegistration;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.esb.platform.base.management.core.ManagementCore;
import ru.entaxy.platform.core.management.object.EntaxyRuntimeItemMBean;

public class MBeanHolder<T extends EntaxyRuntimeItemMBean> {

    private static final Logger log = LoggerFactory.getLogger(MBeanHolder.class);

    public static <S extends EntaxyRuntimeItemMBean> MBeanHolder<S> create(S mbean) {
        MBeanHolder<S> result = new MBeanHolder<S>();
        result.mbean = mbean;
        result.objectId = mbean.getId();
        return result;
    }

    public static <S extends EntaxyRuntimeItemMBean> MBeanHolder<S> register(Class<? extends S> mbeanClass,
            S mbean,
            Dictionary<String, Object> properties,
            BundleContext bundleContext) {

        MBeanHolder<S> result = create(mbean);

        if (properties != null) {
            result.clearProperties();
            Enumeration<String> keys = properties.keys();
            while (keys.hasMoreElements()) {
                String key = keys.nextElement();
                result.registrationProperties.put(key, properties.get(key));
            }
        }

        if (mbeanClass != null)
            result.register(mbeanClass);
        else
            result.register();

        return result;
    }

    ServiceRegistration<?> registration;

    public String objectId;

    public T mbean;

    Dictionary<String, Object> registrationPropertiesDefaults = new Hashtable<>();

    Dictionary<String, Object> registrationProperties = new Hashtable<>();

    BundleContext bundleContext;

    public void clearProperties() {
        this.registrationProperties = new Hashtable<>();
    }

    protected void loadDefaultProperties() {
        Iterator<String> iter = registrationPropertiesDefaults.keys().asIterator();
        while (iter.hasNext()) {
            String key = iter.next();
            registrationProperties.put(key, registrationPropertiesDefaults.get(key));
        }
    }

    public void register() {
        if (this.mbean == null)
            return;
        register(this.mbean.getMBeanClass());
    }

    public void register(Class<? extends EntaxyRuntimeItemMBean> mbeanClass) {
        if (this.mbean == null)
            return;

        unregister();

        String[] classes = new String[] {
            mbeanClass.getName(),
            DynamicMBean.class.getName(),
            MBeanRegistration.class.getName()
        };

        loadDefaultProperties();

        if (registrationProperties.get(ManagementCore.JMX_OBJECTNAME) == null) {
            registrationProperties.put(ManagementCore.JMX_OBJECTNAME,
                    MBeanUtils.addNamespace(MBeanUtils.calculateJmxObjectName(mbean)));
        }

        try {
            registration = bundleContext.registerService(classes, mbean, registrationProperties);
        } catch (Exception e) {
            log.error("Error registering MBean [" + mbean.getId() + "]", e);
        }

    }


    public void unregister() {
        if (registration != null) {
            try {
                registration.unregister();
            } catch (Exception ignore) {
                // NOOP
            }

            registration = null;

        }
        clearProperties();
    }
}
