/*-
 * ~~~~~~licensing~~~~~~
 * object-runtime-core
 * ==========
 * Copyright (C) 2020 - 2025 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.objects.runtime.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.FrameworkUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.platform.base.objects.EntaxyObject;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeObject;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeObjectContainer;

public abstract class EntaxyRuntimeObjectContainerImpl implements EntaxyRuntimeObjectContainer {

    private static final Logger log = LoggerFactory.getLogger(EntaxyRuntimeObjectContainerImpl.class);

    String id;
    String name;

    Map<String, EntaxyRuntimeObjectImpl> objects = new HashMap<>();

    List<EntaxyRuntimeObject> objectsCache = new ArrayList<>();

    Object objectsLock = new Object();

    public void bundle(Bundle bundle) {
        setId(bundle.getBundleId() + "");
        setName(bundle.getSymbolicName());
    }

    public void bundle(long bundleId) {
        bundle(bundleId, FrameworkUtil.getBundle(this.getClass()).getBundleContext());
    }

    public void bundle(long bundleId, BundleContext bundleContext) {
        if (bundleId > 0) {
            // try to find BundleContext that is not null
            // 'cause when first starting compact assembly
            // current context can be null due to refreshing
            BundleContext context = bundleContext;

            if (context == null)
                context = FrameworkUtil.getBundle(getClass()).getBundleContext();
            if (context == null)
                context = FrameworkUtil.getBundle(EntaxyObject.class).getBundleContext();
            if (context != null) {
                Bundle b = context.getBundle(bundleId);
                if (b != null)
                    bundle(b);
            } else {
                log.error("NOT FOUND suitable BundleContext while creating container for [{}]", bundleId);
            }
        }

    }

    public void add(EntaxyRuntimeObjectImpl object) {
        synchronized (objectsLock) {
            objects.put(object.getObjectFullId(), object);
            object.setContainerImpl(this);
            rebuildCache();
        }
    }

    public void remove(EntaxyRuntimeObjectImpl object) {
        synchronized (objectsLock) {
            objects.remove(object.getObjectFullId());
            object.setContainerImpl(null);
            rebuildCache();
        }
    }

    protected void rebuildCache() {
        objectsCache.clear();
        objectsCache.addAll(objects.values().stream().map(v -> (EntaxyRuntimeObject) v).collect(Collectors.toList()));
    }

    @Override
    public String getId() {
        return id;
    }

    @Override
    public String getName() {
        return name;
    }

    public void setId(String id) {
        this.id = id;
    }

    public void setName(String name) {
        this.name = name;
    }

    @Override
    public List<EntaxyRuntimeObject> getObjects() {
        return objectsCache;
    }

    @Override
    public List<EntaxyRuntimeObjectContainer> getParentContainers() {
        final Set<EntaxyRuntimeObjectContainer> containers = new HashSet<>();
        for (EntaxyRuntimeObjectImpl roi : this.objects.values()) {
            roi.getIncomingRelations().stream()
                    .filter(r -> r.isSubordinating())
                    .map(r -> r.getSource())
                    .map(obj -> EntaxyRuntimeObjectServiceImpl.getInstance()
                            .getContainerForObject(obj.getObjectFullId()))
                    .filter(cont -> (cont != null))
                    .filter(cnt -> !this.id.equals(cnt.getId()))
                    .forEach(cont -> containers.add(cont));
        }
        return new ArrayList<>(containers);
    }

    @Override
    public List<EntaxyRuntimeObjectContainer> getChildContainers() {
        final Set<EntaxyRuntimeObjectContainer> containers = new HashSet<>();
        for (EntaxyRuntimeObjectImpl roi : this.objects.values()) {
            roi.getRelations().stream()
                    .filter(r -> r.isSubordinating())
                    .map(r -> r.getTarget())
                    .map(obj -> EntaxyRuntimeObjectServiceImpl.getInstance()
                            .getContainerForObject(obj.getObjectFullId()))
                    .filter(cont -> (cont != null))
                    .filter(cnt -> !this.id.equals(cnt.getId()))
                    .forEach(cont -> containers.add(cont));
        }
        return new ArrayList<>(containers);
    }

    @Override
    public List<EntaxyRuntimeObjectContainer> getDependencies() {
        final Set<EntaxyRuntimeObjectContainer> containers = new HashSet<>();
        for (EntaxyRuntimeObjectImpl roi : this.objects.values()) {
            roi.getRelations().stream()
                    .filter(r -> EntaxyObject.OBJECT_RELATION.DEPENDENCY.equals(r.getType()))
                    .map(r -> r.getTarget())
                    .map(obj -> EntaxyRuntimeObjectServiceImpl.getInstance()
                            .getContainerForObject(obj.getObjectFullId()))
                    .filter(cont -> (cont != null))
                    .filter(cnt -> !this.id.equals(cnt.getId()))
                    .forEach(cont -> containers.add(cont));
        }
        return new ArrayList<>(containers);
    }

    @Override
    public List<EntaxyRuntimeObjectContainer> getDependent() {
        final Set<EntaxyRuntimeObjectContainer> containers = new HashSet<>();
        for (EntaxyRuntimeObjectImpl roi : this.objects.values()) {
            roi.getIncomingRelations().stream()
                    .filter(r -> EntaxyObject.OBJECT_RELATION.DEPENDENCY.equals(r.getType()))
                    .map(r -> r.getSource())
                    .map(obj -> EntaxyRuntimeObjectServiceImpl.getInstance()
                            .getContainerForObject(obj.getObjectFullId()))
                    .filter(cont -> (cont != null))
                    .filter(cnt -> !this.id.equals(cnt.getId()))
                    .forEach(cont -> containers.add(cont));
        }
        return new ArrayList<>(containers);
    }

    @Override
    public List<EntaxyRuntimeObjectContainer> getRelatedContainers() {
        final Set<EntaxyRuntimeObjectContainer> containers = new HashSet<>();
        for (EntaxyRuntimeObjectImpl roi : this.objects.values()) {
            roi.getRelations().stream()
                    .filter(r -> EntaxyObject.OBJECT_RELATION.DEPENDENCY.equals(r.getType()))
                    .map(r -> r.getTarget())
                    .map(obj -> EntaxyRuntimeObjectServiceImpl.getInstance()
                            .getContainerForObject(obj.getObjectFullId()))
                    .filter(cont -> (cont != null))
                    .filter(cnt -> !this.id.equals(cnt.getId()))
                    .forEach(cont -> containers.add(cont));
        }
        return new ArrayList<>(containers);
    }

    @Override
    public List<EntaxyRuntimeObjectContainer> getIncomingRelatedContainers() {
        final Set<EntaxyRuntimeObjectContainer> containers = new HashSet<>();
        for (EntaxyRuntimeObjectImpl roi : this.objects.values()) {
            roi.getIncomingRelations().stream()
                    .filter(r -> EntaxyObject.OBJECT_RELATION.DEPENDENCY.equals(r.getType()))
                    .map(r -> r.getSource())
                    .map(obj -> EntaxyRuntimeObjectServiceImpl.getInstance()
                            .getContainerForObject(obj.getObjectFullId()))
                    .filter(cont -> (cont != null))
                    .filter(cnt -> !this.id.equals(cnt.getId()))
                    .forEach(cont -> containers.add(cont));
        }
        return new ArrayList<>(containers);
    }

}
